@@ -124,7 +124,7 @@ canonicalize_to_upper(char c)
 static apr_size_t
 dirent_root_length(const char *dirent, apr_size_t len)
 {
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
 #ifdef SVN_USE_DOS_PATHS
   if (len >= 2 && dirent[1] == ':' &&
       ((dirent[0] >= 'A' && dirent[0] <= 'Z') ||
@@ -440,7 +440,7 @@ canonicalize(path_type_t type, const char *path, apr_pool_t *pool)
         {
           *(dst++) = *(src++);
 
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
           /* On Windows permit two leading separator characters which means an
            * UNC path. */
           if ((type == type_dirent) && *src == '/')
@@ -497,14 +497,20 @@ canonicalize(path_type_t type, const char *path, apr_pool_t *pool)
 #ifdef SVN_USE_DOS_PATHS
       /* If this is the first path segment of a file:// URI and it contains a
          windows drive letter, convert the drive letter to upper case. */
-      else if (url && canon_segments == 1 && seglen == 2 &&
+      else if (url && canon_segments == 1 && seglen >= 2 &&
                (strncmp(canon, "file:", 5) == 0) &&
                src[0] >= 'a' && src[0] <= 'z' && src[1] == ':')
         {
           *(dst++) = canonicalize_to_upper(src[0]);
           *(dst++) = ':';
-          if (*next)
-            *(dst++) = *next;
+          if (seglen > 2) /* drive relative path */
+            {
+              memcpy(dst, src + 2, seglen - 2);
+              dst += seglen - 2;
+            }
+
+          if (slash_len)
+            *(dst++) = '/';
           canon_segments++;
         }
 #endif /* SVN_USE_DOS_PATHS */
@@ -536,7 +542,7 @@ canonicalize(path_type_t type, const char *path, apr_pool_t *pool)
 
   *dst = '\0';
 
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
   /* Skip leading double slashes when there are less than 2
    * canon segments. UNC paths *MUST* have two segments. */
   if ((type == type_dirent) && canon[0] == '/' && canon[1] == '/')
@@ -670,7 +676,7 @@ get_longest_ancestor_length(path_type_t types,
   apr_size_t path1_len, path2_len;
   apr_size_t i = 0;
   apr_size_t last_dirsep = 0;
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
   svn_boolean_t unc = FALSE;
 #endif
 
@@ -703,7 +709,7 @@ get_longest_ancestor_length(path_type_t types,
     return 0;
 
   /* Handle some windows specific cases */
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
   if (types == type_dirent)
     {
       /* don't count the '//' from UNC paths */
@@ -733,7 +739,7 @@ get_longest_ancestor_length(path_type_t types,
           return i;
 #endif /* SVN_USE_DOS_PATHS */
     }
-#endif /* defined(SVN_USE_DOS_PATHS) || defined(CYGWIN) */
+#endif /* defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__) */
 
   /* last_dirsep is now the offset of the last directory separator we
      crossed before reaching a non-matching byte.  i is the offset of
@@ -748,7 +754,7 @@ get_longest_ancestor_length(path_type_t types,
     {
       /* Nothing in common but the root folder '/' or 'X:/' for Windows
          dirents. */
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
       if (! unc)
         {
 #ifdef SVN_USE_DOS_PATHS
@@ -758,10 +764,10 @@ get_longest_ancestor_length(path_type_t types,
                                && path2[1] == ':' && path2[2] == '/')
             return 3;
 #endif /* SVN_USE_DOS_PATHS */
-#endif /* defined(SVN_USE_DOS_PATHS) || defined(CYGWIN) */
+#endif /* defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__) */
           if (last_dirsep == 0 && path1[0] == '/' && path2[0] == '/')
             return 1;
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
         }
 #endif
     }
@@ -1682,7 +1688,7 @@ svn_dirent_is_canonical(const char *dirent, apr_pool_t *scratch_pool)
   if (*ptr == '/')
     {
       ptr++;
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
       /* Check for UNC paths */
       if (*ptr == '/')
         {
@@ -1859,18 +1865,18 @@ svn_uri_is_canonical(const char *uri, apr_pool_t *scratch_pool)
           ptr++;
         }
 
-      if (ptr == schema_data)
+      if (ptr == schema_data && (*ptr == '/' || *ptr == '\0'))
         return FALSE; /* Fail on "http://host:" */
 
-      if (*ptr && *ptr != '/')
-        return FALSE; /* Not a port number */
-
       if (port == 80 && strncmp(uri, "http:", 5) == 0)
         return FALSE;
       else if (port == 443 && strncmp(uri, "https:", 6) == 0)
         return FALSE;
       else if (port == 3690 && strncmp(uri, "svn:", 4) == 0)
         return FALSE;
+
+      while (*ptr && *ptr != '/')
+        ++ptr; /* Allow "http://host:stuff" */
     }
 
   schema_data = ptr;
@@ -2369,7 +2375,7 @@ svn_uri_get_dirent_from_file_url(const char **dirent,
 
   /* Duplicate the URL, starting at the top of the path.
      At the same time, we URI-decode the path. */
-#if defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)
+#if defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)
   /* On Windows, we'll typically have to skip the leading / if the
      path starts with a drive letter.  Like most Web browsers, We
      support two variants of this scheme:
@@ -2402,8 +2408,11 @@ svn_uri_get_dirent_from_file_url(const char **dirent,
         if (dup_path[1] == '|')
           dup_path[1] = ':';
 
-        if (dup_path[2] == '/' || dup_path[2] == '\0')
+        if (dup_path[2] == '/' || dup_path[2] == '\\' || dup_path[2] == '\0')
           {
+            /* Dirents have upper case drive letters in their canonical form */
+            dup_path[0] = canonicalize_to_upper(dup_path[0]);
+
             if (dup_path[2] == '\0')
               {
                 /* A valid dirent for the driveroot must be like "C:/" instead of
@@ -2416,9 +2425,11 @@ svn_uri_get_dirent_from_file_url(const char **dirent,
                 new_path[3] = '\0';
                 dup_path = new_path;
               }
+            else
+              dup_path[2] = '/'; /* Ensure not relative for '\' after drive! */
           }
       }
-#endif
+#endif    
     if (hostname)
       {
         if (dup_path[0] == '/' && dup_path[1] == '\0')
@@ -2432,7 +2443,7 @@ svn_uri_get_dirent_from_file_url(const char **dirent,
     else
       *dirent = dup_path;
   }
-#else /* !(defined(SVN_USE_DOS_PATHS) || defined(CYGWIN)) */
+#else /* !(defined(SVN_USE_DOS_PATHS) || defined(__CYGWIN__)) */
   /* Currently, the only hostnames we are allowing on non-Win32 platforms
      are the empty string and 'localhost'. */
   if (hostname)
@@ -2460,7 +2471,7 @@ svn_uri_get_file_url_from_dirent(const char **url,
   if (dirent[0] == '/' && dirent[1] == '\0')
     dirent = NULL; /* "file://" is the canonical form of "file:///" */
 
-#ifdef CYGWIN
+#ifdef __CYGWIN__
   if (dirent != NULL && dirent[0] == '/' && dirent[1] == '/')
       /* Handle UNC paths //server/share -> file://server/share */
       *url = apr_pstrcat(pool, "file:", dirent, NULL);
